# Copyright 2022 Inspur
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Thread-safe connection pool for python-binary-memcached."""

from typing import Any
import warnings

try:
    import eventlet  # type: ignore[import-not-found]
except ImportError:
    eventlet = None
import bmemcached
from oslo_cache._memcache_pool import MemcacheClientPool
from oslo_log import log

LOG = log.getLogger(__name__)


if eventlet and eventlet.patcher.is_monkey_patched('thread'):
    warnings.warn(
        "Eventlet support is deprecated and will be removed.",
        category=DeprecationWarning,
        stacklevel=3,
    )


class _BMemcacheClient(bmemcached.Client):  # type: ignore
    """Thread global memcache client

    As client is inherited from threading.local we have to restore object
    methods overloaded by threading.local so we can reuse clients in
    different threads
    """

    __delattr__ = object.__delattr__
    __getattribute__ = object.__getattribute__
    __setattr__ = object.__setattr__

    # Hack for lp 1812935
    if eventlet and eventlet.patcher.is_monkey_patched('thread'):
        # NOTE(bnemec): I'm not entirely sure why this works in a
        # monkey-patched environment and not with vanilla stdlib, but it does.
        def __new__(
            cls, *args: Any, **kwargs: Any
        ) -> type['_BMemcacheClient']:
            return object.__new__(cls)
    else:
        __new__ = object.__new__

    def __del__(self) -> None:
        pass


class BMemcacheClientPool(MemcacheClientPool):
    def __init__(
        self,
        urls: list[str],
        arguments: dict[str, Any],
        *,
        maxsize: int,
        unused_timeout: float,
        conn_get_timeout: float | None = None,
    ) -> None:
        super().__init__(
            urls,
            arguments,
            maxsize=maxsize,
            unused_timeout=unused_timeout,
            conn_get_timeout=conn_get_timeout,
        )
        self._arguments = {
            'username': arguments.get('username', None),
            'password': arguments.get('password', None),
            'tls_context': arguments.get('tls_context', None),
        }

    def _create_connection(self) -> _BMemcacheClient:
        return _BMemcacheClient(self.urls, **self._arguments)



# Copyright 2012 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from collections.abc import Sequence

from oslo_config import cfg


def _supported_backends_msg(backends: list[str]) -> str:
    backends_in_msg = [f'``{name}``' for name in backends]
    return ' '.join(
        [
            'Currently supported by',
            ', '.join(backends_in_msg[:-1]),
            'and',
            backends_in_msg[-1],
        ]
    )


_DEFAULT_BACKEND = 'dogpile.cache.null'

OPTS = [
    cfg.StrOpt(
        'config_prefix',
        default='cache.oslo',
        help=(
            'Prefix for building the configuration dictionary '
            'for the cache region. This should not need to be '
            'changed unless there is another dogpile.cache '
            'region with the same configuration name.'
        ),
    ),
    cfg.IntOpt(
        'expiration_time',
        default=600,
        min=1,
        help=(
            'Default TTL, in seconds, for any cached item in '
            'the dogpile.cache region. This applies to any '
            'cached method that doesn\'t have an explicit '
            'cache expiration time defined for it.'
        ),
    ),
    cfg.IntOpt(
        'backend_expiration_time',
        min=1,
        help=(
            'Expiration time in cache backend to purge '
            'expired records automatically. This should be '
            'greater than expiration_time and all cache_time '
            'options'
        ),
    ),
    # NOTE(morganfainberg): It is recommended that either Redis or
    # Memcached are used as the dogpile backend for real workloads. To
    # prevent issues with the memory cache ending up in "production"
    # unintentionally, we register a no-op as the default caching backend.
    cfg.StrOpt(
        'backend',
        default=_DEFAULT_BACKEND,
        choices=[
            'oslo_cache.memcache_pool',
            'oslo_cache.dict',
            'oslo_cache.etcd3gw',
            'dogpile.cache.pymemcache',
            'dogpile.cache.memcached',
            'dogpile.cache.pylibmc',
            'dogpile.cache.bmemcached',
            'dogpile.cache.dbm',
            'dogpile.cache.redis',
            'dogpile.cache.redis_sentinel',
            'dogpile.cache.memory',
            'dogpile.cache.memory_pickle',
            'dogpile.cache.null',
        ],
        help=(
            'Cache backend module. For eventlet-based or '
            'environments with hundreds of threaded servers, Memcache '
            'with pooling (oslo_cache.memcache_pool) is recommended. '
            'For environments with less than 100 threaded servers, '
            'Memcached (dogpile.cache.memcached) or Redis '
            '(dogpile.cache.redis) is recommended. Test environments '
            'with a single instance of the server can use the '
            'dogpile.cache.memory backend.'
        ),
    ),
    cfg.MultiStrOpt(
        'backend_argument',
        default=[],
        secret=True,
        help=(
            'Arguments supplied to the backend module. '
            'Specify this option once per argument to be '
            'passed to the dogpile.cache backend. Example '
            'format: "<argname>:<value>".'
        ),
    ),
    cfg.ListOpt(
        'proxies',
        default=[],
        help=(
            'Proxy classes to import that will affect the way '
            'the dogpile.cache backend functions. See the '
            'dogpile.cache documentation on '
            'changing-backend-behavior.'
        ),
    ),
    cfg.BoolOpt('enabled', default=False, help='Global toggle for caching.'),
    cfg.BoolOpt(
        'debug_cache_backend',
        default=False,
        help=(
            'Extra debugging from the cache backend (cache '
            'keys, get/set/delete/etc calls). This is only '
            'really useful if you need to see the specific '
            'cache-backend get/set/delete calls with the '
            'keys/values.  Typically this should be left set '
            'to false.'
        ),
    ),
    cfg.ListOpt(
        'memcache_servers',
        default=['localhost:11211'],
        help=(
            'Memcache servers in the format of "host:port". '
            'This is used by backends dependent on Memcached.'
            'If ``dogpile.cache.memcached`` or '
            '``oslo_cache.memcache_pool`` is used and a given '
            'host refer to an IPv6 or a given domain refer to '
            'IPv6 then you should prefix the given address with '
            'the address family (``inet6``) '
            '(e.g ``inet6:[::1]:11211``, '
            '``inet6:[fd12:3456:789a:1::1]:11211``, '
            '``inet6:[controller-0.internalapi]:11211``). '
            'If the address family is not given then these '
            'backends will use the default ``inet`` address '
            'family which corresponds to IPv4'
        ),
    ),
    cfg.IntOpt(
        'memcache_dead_retry',
        default=5 * 60,
        help=(
            'Number of seconds memcached server is considered dead '
            'before it is tried again. (dogpile.cache.memcache and '
            'oslo_cache.memcache_pool backends only).'
        ),
    ),
    cfg.IntOpt(
        'memcache_pool_maxsize',
        default=10,
        help=(
            'Max total number of open connections to every '
            'memcached server. (oslo_cache.memcache_pool backend '
            'only).'
        ),
    ),
    cfg.IntOpt(
        'memcache_pool_unused_timeout',
        default=60,
        help=(
            'Number of seconds a connection to memcached is held '
            'unused in the pool before it is closed. '
            '(oslo_cache.memcache_pool backend only).'
        ),
    ),
    cfg.IntOpt(
        'memcache_pool_connection_get_timeout',
        default=10,
        help=(
            'Number of seconds that an operation will wait to get '
            'a memcache client connection.'
        ),
    ),
    cfg.BoolOpt(
        'memcache_pool_flush_on_reconnect',
        default=False,
        help=(
            'Global toggle if memcache will be flushed '
            'on reconnect. '
            '(oslo_cache.memcache_pool backend only).'
        ),
    ),
    cfg.BoolOpt(
        'memcache_sasl_enabled',
        default=False,
        help=(
            'Enable the SASL(Simple Authentication and Security'
            'Layer) if the SASL_enable is true, else disable.'
        ),
    ),
    cfg.StrOpt(
        'redis_server',
        default='localhost:6379',
        help='Redis server in the format of "host:port"',
    ),
    cfg.IntOpt(
        'redis_db', default=0, min=0, help='Database id in Redis server'
    ),
    cfg.ListOpt(
        'redis_sentinels',
        default=['localhost:26379'],
        help='Redis sentinel servers in the format of "host:port"',
    ),
    cfg.StrOpt(
        'redis_sentinel_service_name',
        default='mymaster',
        help='Service name of the redis sentinel cluster.',
    ),
    cfg.StrOpt(
        'username',
        help='the user name for authentication to backend.',
        deprecated_opts=[
            cfg.DeprecatedOpt('memcache_username', group='cache'),
            cfg.DeprecatedOpt('redis_username', group='cache'),
        ],
    ),
    cfg.StrOpt(
        'password',
        secret=True,
        help='the password for authentication to backend.',
        deprecated_opts=[
            cfg.DeprecatedOpt('memcache_password', group='cache'),
            cfg.DeprecatedOpt('redis_password', group='cache'),
        ],
    ),
    cfg.BoolOpt(
        'tls_enabled',
        default=False,
        help=(
            'Global toggle for TLS usage when communicating with'
            ' the caching servers.'
            + _supported_backends_msg(
                [
                    'dogpile.cache.bmemcache',
                    'dogpile.cache.pymemcache',
                    'oslo_cache.memcache_pool',
                    'dogpile.cache.redis',
                    'dogpile.cache.redis_sentinel',
                ]
            )
        ),
    ),
    cfg.StrOpt(
        'tls_cafile',
        default=None,
        help=(
            'Path to a file of concatenated CA certificates in PEM '
            'format necessary to establish the caching servers\' '
            'authenticity. If tls_enabled is False, this option is '
            'ignored.'
        ),
    ),
    cfg.StrOpt(
        'tls_certfile',
        default=None,
        help=(
            'Path to a single file in PEM format containing the '
            'client\'s certificate as well as any number of CA '
            'certificates needed to establish the certificate\'s '
            'authenticity. This file is only required when client side '
            'authentication is necessary. If tls_enabled is False, '
            'this option is ignored.'
        ),
    ),
    cfg.StrOpt(
        'tls_keyfile',
        default=None,
        help=(
            'Path to a single file containing the client\'s '
            'private key in. Otherwise the private key will be taken '
            'from the file specified in tls_certfile. If tls_enabled '
            'is False, this option is ignored.'
        ),
    ),
    cfg.StrOpt(
        'tls_allowed_ciphers',
        default=None,
        help=(
            'Set the available ciphers for sockets created with '
            'the TLS context. It should be a string in the OpenSSL '
            'cipher list format. If not specified, all OpenSSL enabled '
            'ciphers will be available. '
            + _supported_backends_msg(
                [
                    'dogpile.cache.bmemcache',
                    'dogpile.cache.pymemcache',
                    'oslo_cache.memcache_pool',
                    'dogpile.cache.redis',
                    'dogpile.cache.redis_sentinel',
                ]
            )
        ),
    ),
    cfg.FloatOpt(
        'socket_timeout',
        default=1.0,
        deprecated_opts=[
            cfg.DeprecatedOpt('memcache_socket_timeout', group='cache'),
            cfg.DeprecatedOpt('redis_socket_timeout', group='cache'),
        ],
        help=(
            'Timeout in seconds for every call to a server. '
            + _supported_backends_msg(
                [
                    'dogpile.cache.memcache',
                    'oslo_cache.memcache_pool',
                    'dogpile.cache.redis',
                    'dogpile.cache.redis_sentinel',
                ]
            )
        ),
    ),
    cfg.BoolOpt(
        'enable_socket_keepalive',
        default=False,
        help=(
            "Global toggle for the socket keepalive of "
            "dogpile's pymemcache backend"
        ),
    ),
    cfg.IntOpt(
        'socket_keepalive_idle',
        default=1,
        min=0,
        help=(
            'The time (in seconds) the connection needs to '
            'remain idle before TCP starts sending keepalive probes. '
            'Should be a positive integer most greater than zero.'
        ),
    ),
    cfg.IntOpt(
        'socket_keepalive_interval',
        default=1,
        min=0,
        help=(
            'The time (in seconds) between individual keepalive '
            'probes. Should be a positive integer greater '
            'than zero.'
        ),
    ),
    cfg.IntOpt(
        'socket_keepalive_count',
        default=1,
        min=0,
        help=(
            'The maximum number of keepalive probes TCP should '
            'send before dropping the connection. Should be a '
            'positive integer greater than zero.'
        ),
    ),
    cfg.BoolOpt(
        'enable_retry_client',
        default=False,
        help=(
            'Enable retry client mechanisms to handle failure. '
            'Those mechanisms can be used to wrap all kind of pymemcache '
            'clients. The wrapper allows you to define how many attempts '
            'to make and how long to wait between attempts.'
        ),
    ),
    cfg.IntOpt(
        'retry_attempts',
        min=1,
        default=2,
        help='Number of times to attempt an action before failing.',
    ),
    cfg.FloatOpt(
        'retry_delay',
        default=0,
        help='Number of seconds to sleep between each attempt.',
    ),
    cfg.IntOpt(
        'hashclient_retry_attempts',
        min=1,
        default=2,
        help=(
            'Amount of times a client should be tried '
            'before it is marked dead and removed from the pool in '
            'the HashClient\'s internal mechanisms.'
        ),
    ),
    cfg.FloatOpt(
        'hashclient_retry_timeout',
        deprecated_name='hashclient_retry_delay',
        default=1,
        help=(
            'Time in seconds that should pass between '
            'retry attempts in the HashClient\'s internal mechanisms.'
        ),
    ),
    cfg.FloatOpt(
        'hashclient_dead_timeout',
        deprecated_name='dead_timeout',
        default=60,
        help=(
            'Time in seconds before attempting to add a node '
            'back in the pool in the HashClient\'s internal mechanisms.'
        ),
    ),
]


def configure(conf: cfg.ConfigOpts) -> None:
    conf.register_opts(OPTS, 'cache')


def set_defaults(
    conf: cfg.ConfigOpts, memcache_pool_flush_on_reconnect: bool = False
) -> None:
    """Set defaults for configuration variables.

    Overrides default options values.

    :param conf: Configuration object, managed by the caller.
    :type conf: oslo.config.cfg.ConfigOpts

    :param memcache_pool_flush_on_reconnect: The default state for the
        ``flush_on_reconnect`` flag. By default deactivated
    :type memcache_pool_flush_on_reconnect: bool
    """
    conf.register_opt(OPTS, group='cache')

    cfg.set_defaults(
        OPTS,
        memcache_pool_flush_on_reconnect=memcache_pool_flush_on_reconnect,
    )


def list_opts() -> list[tuple[str, Sequence[cfg.Opt]]]:
    """Return a list of oslo_config options.

    The returned list includes all oslo_config options which are registered as
    the "OPTS".

    Each object in the list is a two element tuple. The first element of
    each tuple is the name of the group under which the list of options in the
    second element will be registered. A group name of None corresponds to the
    [DEFAULT] group in config files.

    This function is also discoverable via the 'oslo_config.opts' entry point
    under the 'oslo_cache.config.opts' namespace.

    The purpose of this is to allow tools like the Oslo sample config file
    generator to discover the options exposed to users by this library.

    :returns: a list of (group_name, opts) tuples
    """
    return [('cache', OPTS)]



# Copyright 2015 Mirantis Inc
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""dogpile.cache backend that uses etcd 3.x for storage"""

from collections.abc import Iterable, Mapping, Sequence
from typing import Any

from dogpile.cache import api

from oslo_cache import core
from oslo_serialization import jsonutils

__all__ = ['Etcd3gwCacheBackend']

_NO_VALUE = core.NO_VALUE


class Etcd3gwCacheBackend(api.CacheBackend):
    #: Default socket/lock/member/leader timeout used when none is provided.
    DEFAULT_TIMEOUT = 30

    #: Default hostname used when none is provided.
    DEFAULT_HOST = "localhost"

    #: Default port used if none provided (4001 or 2379 are the common ones).
    DEFAULT_PORT = 2379

    def __init__(self, arguments: dict[str, Any]) -> None:
        self.host = arguments.get('host', self.DEFAULT_HOST)
        self.port = arguments.get('port', self.DEFAULT_PORT)
        self.timeout = int(arguments.get('timeout', self.DEFAULT_TIMEOUT))
        # module etcd3gw is only required when etcd3gw backend is used
        import etcd3gw

        self._client = etcd3gw.client(
            host=self.host, port=self.port, timeout=self.timeout
        )

    def get(self, key: api.KeyType) -> api.BackendFormatted:
        values = self._client.get(key, False)
        if not values:
            return core.NO_VALUE
        value, metadata = jsonutils.loads(values[0])
        return api.CachedValue(value, metadata)

    def get_multi(
        self, keys: Iterable[api.KeyType]
    ) -> Sequence[api.BackendFormatted]:
        """Retrieves the value for a list of keys."""
        return [self.get(key) for key in keys]

    def set(self, key: api.KeyType, value: api.BackendSetType) -> None:
        self.set_multi({key: value})

    def set_multi(
        self, mapping: Mapping[api.KeyType, api.BackendSetType]
    ) -> None:
        lease = None
        if self.timeout:
            lease = self._client.lease(ttl=self.timeout)
        for key, value in mapping.items():
            self._client.put(key, jsonutils.dumps(value), lease)

    def delete(self, key: api.KeyType) -> None:
        self._client.delete(key)

    def delete_multi(self, keys: Iterable[api.KeyType]) -> None:
        for key in keys:
            self._client.delete(key)



#    Copyright 2021 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

from oslo_cache.tests.functional import test_base


MEMCACHED_PORT = os.getenv("OSLO_CACHE_TEST_MEMCACHED_PORT", "11212")


class TestDogpileCachePyMemcacheBackend(test_base.BaseTestCaseCacheBackend):
    def setUp(self):
        self.config_fixture.config(
            group="cache",
            backend="dogpile.cache.pymemcache",
            memcache_servers=[f'localhost:{MEMCACHED_PORT}'],
        )

        # NOTE(hberaud): super must be called after all to ensure that
        # config fixture is properly initialized with value related to
        # the current backend in use.
        super().setUp()



# Copyright 2013 Metacloud
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Items useful for external testing."""

import copy
from typing import TypeVar

from dogpile.cache import api
from dogpile.cache import proxy

from oslo_cache import core as cache


__all__ = [
    'CacheIsolatingProxy',
]


NO_VALUE = cache.NO_VALUE


_T = TypeVar('_T')


def _copy_value(value: _T) -> _T:
    if value is not NO_VALUE:
        value = copy.deepcopy(value)
    return value


# NOTE(morganfainberg): WARNING - It is not recommended to use the Memory
# backend for dogpile.cache in a real deployment under any circumstances. The
# backend does no cleanup of expired values and therefore will leak memory. The
# backend is not implemented in a way to share data across processes (e.g.
# Keystone in HTTPD.  This proxy is a hack to get around the lack of isolation
# of values in memory.  Currently it blindly stores and retrieves the values
# from the cache, and modifications to dicts/lists/etc returned can result in
# changes to the cached values.  In short, do not use the dogpile.cache.memory
# backend unless you are running tests or expecting odd/strange results.
class CacheIsolatingProxy(proxy.ProxyBackend):
    """Proxy that forces a memory copy of stored values.

    The default in-memory cache-region does not perform a copy on values it
    is meant to cache.  Therefore if the value is modified after set or after
    get, the cached value also is modified.  This proxy does a copy as the last
    thing before storing data.

    In your application's tests, you'll want to set this as a proxy for the
    in-memory cache, like this::

        self.config_fixture.config(
            group='cache',
            backend='dogpile.cache.memory',
            enabled=True,
            proxies=['oslo_cache.testing.CacheIsolatingProxy'],
        )

    """

    def get(self, key: api.KeyType) -> api.BackendFormatted:
        return _copy_value(self.proxied.get(key))

    def set(self, key: api.KeyType, value: api.BackendSetType) -> None:
        self.proxied.set(key, _copy_value(value))



# Copyright 2013 Metacloud
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import fixture as config_fixture
from oslotest import base


class BaseTestCase(base.BaseTestCase):
    def setUp(self):
        super().setUp()
        self.config_fixture = self.useFixture(config_fixture.Config())
        self.config_fixture.config(
            # TODO(morganfainberg): Make Cache Testing a separate test case
            # in tempest, and move it out of the base unit tests.
            group='cache',
            backend='dogpile.cache.memory',
            enabled=True,
            proxies=['oslo_cache.testing.CacheIsolatingProxy'],
        )



# Copyright 2014 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import fixture
from oslo_utils import uuidutils
from oslotest import base

from oslo_cache import core as cache


NO_VALUE = cache.NO_VALUE


class BaseTestCaseCacheBackend(base.BaseTestCase):
    def setUp(self):
        super().setUp()

        self.conf = self.config_fixture.conf

        self.region = cache.create_region()
        self.region_kwargs = cache.create_region(
            function=cache.kwarg_function_key_generator
        )

        cache.configure_cache_region(self.conf, self.region)
        cache.configure_cache_region(self.conf, self.region_kwargs)

    @property
    def config_fixture(self):
        if not hasattr(self, "_config_fixture"):
            self._config_fixture = self.useFixture(fixture.Config())

            # default config for all tests
            self._config_fixture.config(
                group="cache",
                enabled=True,
            )

        return self._config_fixture

    def test_backend_get_missing_data(self):
        random_key = uuidutils.generate_uuid(dashed=False)
        # should return NO_VALUE as key does not exist in cache
        self.assertEqual(NO_VALUE, self.region.get(random_key))

    def test_backend_set_data(self):
        random_key = uuidutils.generate_uuid(dashed=False)
        self.region.set(random_key, "dummyValue")
        self.assertEqual("dummyValue", self.region.get(random_key))

    def test_backend_set_none_as_data(self):
        random_key = uuidutils.generate_uuid(dashed=False)
        self.region.set(random_key, None)
        self.assertIsNone(self.region.get(random_key))

    def test_backend_set_blank_as_data(self):
        random_key = uuidutils.generate_uuid(dashed=False)
        self.region.set(random_key, "")
        self.assertEqual("", self.region.get(random_key))

    def test_backend_set_same_key_multiple_times(self):
        random_key = uuidutils.generate_uuid(dashed=False)
        self.region.set(random_key, "dummyValue")
        self.assertEqual("dummyValue", self.region.get(random_key))

        dict_value = {'key1': 'value1'}
        self.region.set(random_key, dict_value)
        self.assertEqual(dict_value, self.region.get(random_key))

        self.region.set(random_key, "dummyValue2")
        self.assertEqual("dummyValue2", self.region.get(random_key))

    def test_backend_multi_set_data(self):
        random_key = uuidutils.generate_uuid(dashed=False)
        random_key1 = uuidutils.generate_uuid(dashed=False)
        random_key2 = uuidutils.generate_uuid(dashed=False)
        random_key3 = uuidutils.generate_uuid(dashed=False)
        mapping = {
            random_key1: 'dummyValue1',
            random_key2: 'dummyValue2',
            random_key3: 'dummyValue3',
        }
        self.region.set_multi(mapping)
        # should return NO_VALUE as key does not exist in cache
        self.assertEqual(NO_VALUE, self.region.get(random_key))
        self.assertFalse(self.region.get(random_key))
        self.assertEqual("dummyValue1", self.region.get(random_key1))
        self.assertEqual("dummyValue2", self.region.get(random_key2))
        self.assertEqual("dummyValue3", self.region.get(random_key3))

    def test_backend_multi_get_data(self):
        random_key = uuidutils.generate_uuid(dashed=False)
        random_key1 = uuidutils.generate_uuid(dashed=False)
        random_key2 = uuidutils.generate_uuid(dashed=False)
        random_key3 = uuidutils.generate_uuid(dashed=False)
        mapping = {
            random_key1: 'dummyValue1',
            random_key2: '',
            random_key3: 'dummyValue3',
        }
        self.region.set_multi(mapping)

        keys = [random_key, random_key1, random_key2, random_key3]
        results = self.region.get_multi(keys)
        # should return NO_VALUE as key does not exist in cache
        self.assertEqual(NO_VALUE, results[0])
        self.assertEqual("dummyValue1", results[1])
        self.assertEqual("", results[2])
        self.assertEqual("dummyValue3", results[3])

    def test_backend_multi_set_should_update_existing(self):
        random_key = uuidutils.generate_uuid(dashed=False)
        random_key1 = uuidutils.generate_uuid(dashed=False)
        random_key2 = uuidutils.generate_uuid(dashed=False)
        random_key3 = uuidutils.generate_uuid(dashed=False)
        mapping = {
            random_key1: 'dummyValue1',
            random_key2: 'dummyValue2',
            random_key3: 'dummyValue3',
        }
        self.region.set_multi(mapping)
        # should return NO_VALUE as key does not exist in cache
        self.assertEqual(NO_VALUE, self.region.get(random_key))
        self.assertEqual("dummyValue1", self.region.get(random_key1))
        self.assertEqual("dummyValue2", self.region.get(random_key2))
        self.assertEqual("dummyValue3", self.region.get(random_key3))

        mapping = {random_key1: 'dummyValue4', random_key2: 'dummyValue5'}
        self.region.set_multi(mapping)
        self.assertEqual(NO_VALUE, self.region.get(random_key))
        self.assertEqual("dummyValue4", self.region.get(random_key1))
        self.assertEqual("dummyValue5", self.region.get(random_key2))
        self.assertEqual("dummyValue3", self.region.get(random_key3))

    def test_backend_multi_set_get_with_blanks_none(self):
        random_key = uuidutils.generate_uuid(dashed=False)
        random_key1 = uuidutils.generate_uuid(dashed=False)
        random_key2 = uuidutils.generate_uuid(dashed=False)
        random_key3 = uuidutils.generate_uuid(dashed=False)
        random_key4 = uuidutils.generate_uuid(dashed=False)
        mapping = {
            random_key1: 'dummyValue1',
            random_key2: None,
            random_key3: '',
            random_key4: 'dummyValue4',
        }
        self.region.set_multi(mapping)
        # should return NO_VALUE as key does not exist in cache
        self.assertEqual(NO_VALUE, self.region.get(random_key))
        self.assertEqual("dummyValue1", self.region.get(random_key1))
        self.assertIsNone(self.region.get(random_key2))
        self.assertEqual("", self.region.get(random_key3))
        self.assertEqual("dummyValue4", self.region.get(random_key4))

        keys = [random_key, random_key1, random_key2, random_key3, random_key4]
        results = self.region.get_multi(keys)

        # should return NO_VALUE as key does not exist in cache
        self.assertEqual(NO_VALUE, results[0])
        self.assertEqual("dummyValue1", results[1])
        self.assertIsNone(results[2])
        self.assertEqual("", results[3])
        self.assertEqual("dummyValue4", results[4])

        mapping = {random_key1: 'dummyValue5', random_key2: 'dummyValue6'}
        self.region.set_multi(mapping)
        self.assertEqual(NO_VALUE, self.region.get(random_key))
        self.assertEqual("dummyValue5", self.region.get(random_key1))
        self.assertEqual("dummyValue6", self.region.get(random_key2))
        self.assertEqual("", self.region.get(random_key3))

    def test_backend_delete_data(self):
        random_key = uuidutils.generate_uuid(dashed=False)
        self.region.set(random_key, "dummyValue")
        self.assertEqual("dummyValue", self.region.get(random_key))

        self.region.delete(random_key)
        # should return NO_VALUE as key no longer exists in cache
        self.assertEqual(NO_VALUE, self.region.get(random_key))

    def test_backend_multi_delete_data(self):
        random_key = uuidutils.generate_uuid(dashed=False)
        random_key1 = uuidutils.generate_uuid(dashed=False)
        random_key2 = uuidutils.generate_uuid(dashed=False)
        random_key3 = uuidutils.generate_uuid(dashed=False)
        mapping = {
            random_key1: 'dummyValue1',
            random_key2: 'dummyValue2',
            random_key3: 'dummyValue3',
        }
        self.region.set_multi(mapping)
        # should return NO_VALUE as key does not exist in cache
        self.assertEqual(NO_VALUE, self.region.get(random_key))
        self.assertEqual("dummyValue1", self.region.get(random_key1))
        self.assertEqual("dummyValue2", self.region.get(random_key2))
        self.assertEqual("dummyValue3", self.region.get(random_key3))
        self.assertEqual(NO_VALUE, self.region.get("InvalidKey"))

        keys = mapping.keys()

        self.region.delete_multi(list(keys))

        self.assertEqual(NO_VALUE, self.region.get("InvalidKey"))
        # should return NO_VALUE as keys no longer exist in cache
        self.assertEqual(NO_VALUE, self.region.get(random_key1))
        self.assertEqual(NO_VALUE, self.region.get(random_key2))
        self.assertEqual(NO_VALUE, self.region.get(random_key3))



# Copyright 2015 Mirantis Inc
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""dogpile.cache backend that uses dictionary for storage"""

from collections.abc import Iterable, Mapping, Sequence
from typing import Any

from dogpile.cache import api
from oslo_cache import core
from oslo_utils import timeutils

__all__ = ['DictCacheBackend']

_NO_VALUE = core.NO_VALUE


class DictCacheBackend(api.CacheBackend):
    """A DictCacheBackend based on dictionary.

    Arguments accepted in the arguments dictionary:

    :param expiration_time: interval in seconds to indicate maximum
        time-to-live value for each key in DictCacheBackend.
        Default expiration_time value is 0, that means that all keys have
        infinite time-to-live value.
    :type expiration_time: real
    """

    def __init__(self, arguments: dict[str, Any]) -> None:
        self.expiration_time = arguments.get('expiration_time', 0)
        self.cache: dict[str, tuple[api.BackendFormatted, float]] = {}

    def get(self, key: api.KeyType) -> api.BackendFormatted:
        """Retrieves the value for a key.

        :param key: dictionary key
        :returns: value for a key or :data:`oslo_cache.core.NO_VALUE`
            for nonexistent or expired keys.
        """
        value, timeout = self.cache.get(key, (_NO_VALUE, 0))
        if self.expiration_time > 0 and timeutils.utcnow_ts() >= timeout:
            self.cache.pop(key, None)
            return _NO_VALUE

        return value

    def get_multi(
        self, keys: Iterable[api.KeyType]
    ) -> Sequence[api.BackendFormatted]:
        """Retrieves the value for a list of keys."""
        return [self.get(key) for key in keys]

    def set(self, key: api.KeyType, value: api.BackendSetType) -> None:
        """Sets the value for a key.

        Expunges expired keys during each set.

        :param key: dictionary key
        :param value: value associated with the key
        """
        self.set_multi({key: value})

    def set_multi(
        self, mapping: Mapping[api.KeyType, api.BackendSetType]
    ) -> None:
        """Set multiple values in the cache.
        Expunges expired keys during each set.

        :param mapping: dictionary with key/value pairs
        """
        self._clear()
        timeout = 0
        if self.expiration_time > 0:
            timeout = timeutils.utcnow_ts() + self.expiration_time
        for key, value in mapping.items():
            self.cache[key] = (value, timeout)

    def delete(self, key: api.KeyType) -> None:
        """Deletes the value associated with the key if it exists.

        :param key: dictionary key
        """
        self.cache.pop(key, None)

    def delete_multi(self, keys: Iterable[api.KeyType]) -> None:
        """Deletes the value associated with each key in list if it exists.

        :param keys: list of dictionary keys
        """
        for key in keys:
            self.cache.pop(key, None)

    def _clear(self) -> None:
        """Expunges expired keys."""
        now = timeutils.utcnow_ts()
        for k in list(self.cache):
            (_value, timeout) = self.cache[k]
            if timeout > 0 and now >= timeout:
                del self.cache[k]
